# Dictionary of available modules. Can be added to anytime and code will still work.
# Would require graph generation methods in FastQCManager if graph output was needed.
module_dict = {"BS": "Basic Statistics",
               1:  "Per base sequence quality",
               2:  "Per tile sequence quality",
               3:  "Per sequence quality scores",
               4:  "Per base sequence content",
               5:  "Per sequence GC content",
               6:  "Per base N content",
               7:  "Sequence Length Distribution",
               8:  "Sequence Duplication Levels",
               9:  "Overrepresented sequences",
               10: "Adapter Content",
               11: "Kmer Content",
               }
